// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file ElementObjectData.h */

#ifndef AMDIS_ELEMENTOBJECTDATA_H
#define AMDIS_ELEMENTOBJECTDATA_H

#include <map>
#include <vector>
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>

#include "Global.h"
#include "Boundary.h"
#include "Serializer.h"

namespace AMDiS {

  using namespace std;

  template<typename T>
  struct PerBoundMap 
  {
    typedef map<pair<T, T>, BoundaryType> type;
    typedef typename type::iterator iterator;
  };

  struct ElementObjectData {
    ElementObjectData(int a = -1, int b = 0, BoundaryType c = INTERIOR)
      : elIndex(a),
	ithObject(b),
	boundaryType(c)
    {}
    
    int elIndex;
    
    int ithObject;
    
    BoundaryType boundaryType;


    void serialize(ostream &out) const
    {
      SerUtil::serialize(out, elIndex);
      SerUtil::serialize(out, ithObject);
      SerUtil::serialize(out, boundaryType);
    }


    void deserialize(istream &in)
    {
      SerUtil::deserialize(in, elIndex);
      SerUtil::deserialize(in, ithObject);
      SerUtil::deserialize(in, boundaryType);
    }


    bool operator==(ElementObjectData& cmp) const
    {
      return (elIndex == cmp.elIndex &&
	      ithObject == cmp.ithObject &&
	      boundaryType == cmp.boundaryType);
    }

    bool operator<(const ElementObjectData& rhs) const
    {
      return (elIndex < rhs.elIndex ||
	      (elIndex == rhs.elIndex &&
	       ithObject < rhs.ithObject));
    }
  };



  class ElementObjects {
  public:
    ElementObjects()
      : iterGeoPos(CENTER)
    {}


    void addVertex(Element *el, int ith, BoundaryType bound = INTERIOR)
    {
      DegreeOfFreedom vertex = el->getDof(ith, 0);
      int elIndex = el->getIndex();
      ElementObjectData elObj(elIndex, ith, bound);

      vertexElements[vertex].push_back(elObj);
      vertexLocalMap[elObj] = vertex;
    }


    void addEdge(Element *el, int ith, BoundaryType bound = INTERIOR)
    {
      DofEdge edge = el->getEdge(ith);
      int elIndex = el->getIndex();
      ElementObjectData elObj(elIndex, ith, bound);

      edgeElements[edge].push_back(elObj);
      edgeLocalMap[elObj] = edge;
    }


    void addFace(Element *el, int ith, BoundaryType bound = INTERIOR)
    {
      DofFace face = el->getFace(ith);
      int elIndex = el->getIndex();
      ElementObjectData elObj(elIndex, ith, bound);

      faceElements[face].push_back(elObj);
      faceLocalMap[elObj] = face;
    }


    void addElement(Element *el, BoundaryType bound = INTERIOR)
    {
      for (int i = 0; i < el->getGeo(VERTEX); i++)
	addVertex(el, i);

      for (int i = 0; i < el->getGeo(EDGE); i++)
	addEdge(el, i);      

      for (int i = 0; i < el->getGeo(FACE); i++)
	addFace(el, i);
    }

    void addPeriodicVertex(pair<DegreeOfFreedom, DegreeOfFreedom> perVertex, 
			   BoundaryType bType)
    {
      periodicVertices[perVertex] = bType;
    }
    
    void addPeriodicEdge(pair<DofEdge, DofEdge> perEdge, BoundaryType bType)
    {
      periodicEdges[perEdge] = bType;
    }

    void addPeriodicFace(pair<DofFace, DofFace> perFace, BoundaryType bType)
    {
      periodicFaces[perFace] = bType;
    }

    void createRankData(map<int, int>& macroElementRankMap);


    bool iterate(GeoIndex pos)
    {
      if (iterGeoPos == CENTER) {
	iterGeoPos = pos;
	switch (iterGeoPos) {
	case VERTEX:
	  vertexIter = vertexInRank.begin();
	  break;
	case EDGE:
	  edgeIter = edgeInRank.begin();
	  break;
	case FACE:
	  faceIter = faceInRank.begin();
	  break;
	default:
	  ERROR_EXIT("Not GeoIndex %d!\n", iterGeoPos);
	}
      } else {
	switch (iterGeoPos) {
	case VERTEX:
	  ++vertexIter;
	  break;
	case EDGE:
	  ++edgeIter;
	  break;
	case FACE:
	  ++faceIter;
	  break;
	default:
	  ERROR_EXIT("Not GeoIndex %d!\n", iterGeoPos);
	}
      }

      switch (iterGeoPos) {
      case VERTEX:
	if (vertexIter == vertexInRank.end()) {
	  iterGeoPos = CENTER;
	  return false;
	}
	break;
      case EDGE:
	if (edgeIter == edgeInRank.end()) {
	  iterGeoPos = CENTER;
	  return false;
	}
	break;
      case FACE:
	if (faceIter == faceInRank.end()) {
	  iterGeoPos = CENTER;
	  return false;
	}
	break;
      default:
	ERROR_EXIT("Should not happen!\n");	
      }

      return true;
    }


    map<int, ElementObjectData>& getIterateData()
    {
      switch (iterGeoPos) {
      case VERTEX:
	return vertexIter->second;
	break;
      case EDGE:
	return edgeIter->second;
	break;
      case FACE:
	return faceIter->second;
	break;
      default:
	ERROR_EXIT("Should not happen!\n");

	// Will never be reached, just to avoid compiler warnings.
	return faceIter->second;
      }
    }


    int getIterateOwner()
    {
      switch (iterGeoPos) {
      case VERTEX:
	return vertexOwner[vertexIter->first];
	break;
      case EDGE:
	return edgeOwner[edgeIter->first];
	break;
      case FACE:
	return faceOwner[faceIter->first];
	break;
      default:
	ERROR_EXIT("Should not happen!\n");

	// Will never be reached, just to avoid compiler warnings.
	return -1;
      }
    }


    int getOwner(DegreeOfFreedom vertex)
    {
      return vertexOwner[vertex];
    }

    int getOwner(DofEdge edge)
    {
      return edgeOwner[edge];
    }

    int getOwner(DofFace face)
    {
      return faceOwner[face];
    }


    int isInRank(DegreeOfFreedom vertex, int rank)
    {
      return (vertexInRank[vertex].count(rank));
    }

    int isInRank(DofEdge edge, int rank)
    {
      return (edgeInRank[edge].count(rank));
    }

    int isInRank(DofFace face, int rank)
    {
      return (faceInRank[face].count(rank));
    }


    vector<ElementObjectData>& getElements(DegreeOfFreedom vertex)
    {
      return vertexElements[vertex];
    }

    vector<ElementObjectData>& getElements(DofEdge edge)
    {
      return edgeElements[edge];
    }

    vector<ElementObjectData>& getElements(DofFace face)
    {
      return faceElements[face];
    }


    map<int, ElementObjectData>& getElementsInRank(DegreeOfFreedom vertex)
    {
      return vertexInRank[vertex];
    }

    map<int, ElementObjectData>& getElementsInRank(DofEdge edge)
    {
      return edgeInRank[edge];
    }

    map<int, ElementObjectData>& getElementsInRank(DofFace face)
    {
      return faceInRank[face];
    }

    DegreeOfFreedom getVertexLocalMap(ElementObjectData &data)
    {
      return vertexLocalMap[data];
    }

    DofEdge getEdgeLocalMap(ElementObjectData &data)
    {
      return edgeLocalMap[data];
    }

    DofFace getFaceLocalMap(ElementObjectData &data)
    {
      return faceLocalMap[data];
    }

    void serialize(ostream &out);

    void deserialize(istream &in);

  private:
    void serialize(ostream &out, vector<ElementObjectData>& elVec);

    void deserialize(istream &in, vector<ElementObjectData>& elVec);

    void serialize(ostream &out, map<int, ElementObjectData>& data);

    void deserialize(istream &in, map<int, ElementObjectData>& data);

  private:
    map<DegreeOfFreedom, vector<ElementObjectData> > vertexElements;
    map<DofEdge, vector<ElementObjectData> > edgeElements;
    map<DofFace, vector<ElementObjectData> > faceElements;


    map<ElementObjectData, DegreeOfFreedom> vertexLocalMap;
    map<ElementObjectData, DofEdge> edgeLocalMap;
    map<ElementObjectData, DofFace> faceLocalMap;


    map<DegreeOfFreedom, int> vertexOwner;
    map<DofEdge, int> edgeOwner;
    map<DofFace, int> faceOwner;


    map<DegreeOfFreedom, map<int, ElementObjectData> > vertexInRank;
    map<DofEdge, map<int, ElementObjectData> > edgeInRank;
    map<DofFace, map<int, ElementObjectData> > faceInRank;

    map<DegreeOfFreedom, map<int, ElementObjectData> >::iterator vertexIter;
    map<DofEdge, map<int, ElementObjectData> >::iterator edgeIter;
    map<DofFace, map<int, ElementObjectData> >::iterator faceIter;

    GeoIndex iterGeoPos;

  public:
    // The following three data structures store periodic DOFs, edges and faces.
    PerBoundMap<DegreeOfFreedom>::type periodicVertices;
    PerBoundMap<DofEdge>::type periodicEdges;
    PerBoundMap<DofFace>::type periodicFaces;
  };

}

#endif
